# -*- coding: utf-8 -*-

import os, subprocess, itertools, time

from tralda.datastructures.Tree import TreeNode

from asymmetree.analysis.BestMatches import bmg_from_tree

from asymmetree.tools.PhyloTreeTools import (delete_and_reconnect,
                                             distances_from_root,
                                             parse_newick,)

from bmgedit.best_match_infer.ScenarioFileIO import matrix_to_phylip


__author__ = 'David Schaller'


# --------------------------------------------------------------------------
#                            TREE RECONSTRUCTION
#
# --------------------------------------------------------------------------

def bmg_by_tree_reconstruction(scenario, matrix_filename,
                               supply_rbmg=True, return_calltime=False):
    
    if return_calltime:
        start_time = time.time()
    
    nj_tree = neighbor_joining(scenario.genes, scenario.gene_index, matrix_filename)
    midpoint_rooting(nj_tree)
    bmg, rbmg = bmg_from_tree(nj_tree, supply_rbmg=supply_rbmg)
    
    if return_calltime:
        return bmg, rbmg, time.time() - start_time
    else:
        return bmg, rbmg
        
        

def neighbor_joining(leaves, leaf_index, matrix_filename,
                     return_calltime=False,
                     binary_path=None):
    """
    Keyword argument:
        return_calltime -- return the time needed for RapidNJ call
        binary_path -- path to 'qinfer' binary (if not available
                       within path)
    """
    
    if not binary_path:
        nj_command = "rapidnj"
    elif os.path.exists(binary_path):
         nj_command = binary_path
    else:
        raise FileNotFoundError("path to RapidNJ binary file '{}' does not "\
                                "exist".format(binary_path))
            
    start_time = time.time()
    
    try:
        output = subprocess.run([nj_command, matrix_filename, "-i", "pd"],
                                stdout=subprocess.PIPE)
    except:
        raise FileNotFoundError("calling RapidNJ failed")
        
    calltime = time.time() - start_time
    
    newick = output.stdout.decode()
    
    tree = parse_newick(newick)
    
    # restore (leaf) indeces and reconciliation
    leaf_dict = {leaf.label: leaf for leaf in leaves}
    index_counter = 0
    for v in tree.preorder():
        if not v.children:
            v.reconc = leaf_dict[v.label].reconc
        else:
            while index_counter in leaf_dict:
                index_counter += 1
            v.label = index_counter
            index_counter += 1
    
    if return_calltime:
        return tree, calltime
    else:
        return tree
    

def reroot(tree, node):
    
    edges = []                              # edges (u, v, distance) to be changed
    pos = node
    while pos.parent:
        edges.append( (pos.parent, pos, pos.dist) )    
        pos = pos.parent
    old_root = pos
    
    for u, v, dist in reversed(edges):      # change direction of edges
        u.remove_child(v)
        v.add_child(u)
        u.dist = dist
    
    node.detach()
    node.dist = 0.0
    tree.root = node
    
    if len(old_root.children) <= 1:         # delete old root if its out-degree is 1
        delete_and_reconnect(tree, old_root)


def midpoint_rooting(tree):
    
    # identify the two most distant leaves and their l.c.a.
    distance_dict = distances_from_root(tree)
    leaf_dict = tree.leaf_dict()
    max_dist, leaf1, leaf2, lca = float('-inf'), None, None, None
    max_label = 0
    
    for v in tree.preorder():
        if v.children:
            for c1, c2 in itertools.combinations(v.children, 2):
                for x in leaf_dict[c1]:
                    x_dist = distance_dict[x] - distance_dict[v]
                    for y in leaf_dict[c2]:
                        y_dist = distance_dict[y] - distance_dict[v]
                        if x_dist + y_dist > max_dist:
                            max_dist, leaf1, leaf2, lca = x_dist + y_dist, x, y, v
        max_label = max([max_label, v.label])
    
    # identify the edge for the new root
    edge, cut_at = None, 0
    
    pos, remaining = leaf1, max_dist/2
    while (not edge) and (pos is not lca):
        if remaining < pos.dist:
            edge = (pos.parent, pos)
            cut_at = pos.dist - remaining
        remaining -= pos.dist
        pos = pos.parent
    
    pos, remaining = leaf2, max_dist/2
    while (not edge) and (pos is not lca):
        if remaining < pos.dist:
            edge = (pos.parent, pos)
            cut_at = pos.dist - remaining
        remaining -= pos.dist
        pos = pos.parent
    
    # rerooting
    if edge:
        u, v = edge
        u.remove_child(v)
        new_root = TreeNode(label=max_label+1, dist=cut_at)
        u.add_child(new_root)
        new_root.add_child(v)
        v.dist -= new_root.dist
        reroot(tree, new_root)
    else:
        reroot(tree, lca)
        

def nj_from_numpy_matrix(leaves, leaf_index, matrix,
                         filename='temp.phylip'):

    matrix_to_phylip(filename, leaves, matrix)
    tree = neighbor_joining(leaves, leaf_index, filename)
    os.remove(filename)
    
    return tree



# -*- coding: utf-8 -*-


import heapq, itertools


__author__ = 'David Schaller'


class _TreeNode:
    
    def __init__(self, priority, value):
        
        self.priority = priority
        self.value = value
        self.children = []
    
    
    def __eq__(self, o):
        
        return self.priority == o.priority
    
    
    def __le__(self, o):
        
        return self.priority <= o.priority
    
    
    def __lt__(self, o):
        
        return self.priority < o.priority


def karmarkar_karp(iterable):
    
    heap = []
    
    for item in iterable:
        if isinstance(item, int):
            heapq.heappush(heap, (_TreeNode(-item, item)))
        else:
            heapq.heappush(heap, (_TreeNode(-len(item), item)))
    
    while len(heap) > 1:
        x = heapq.heappop(heap)
        y = heapq.heappop(heap)
        x.children.append(y)
        x.priority -= y.priority
        heapq.heappush(heap, x)
    
    root = heap[0]
    
    # construct the actual partition via 2-coloring of the tree
    part = [[],[]]
    stack = [(root, 0)]
    
    while stack:
        node, color = stack.pop()
        part[color].append(node.value)
        
        for child in node.children:
            stack.append( (child, (color+1)%2) )
    
    return part


def balanced_coarse_graining(partition):
    
    return [[item for item in itertools.chain.from_iterable(set_i)]
            for set_i in karmarkar_karp(partition)]


if __name__ == '__main__':
    
    numbers = [8, 7, 6, 5, 4]
    print(karmarkar_karp(numbers))
    
    part = [[3,2,1], [4,5,6,7], [8,9], [10,11,12,13,14], [15]]
    print(balanced_coarse_graining(part))


# -*- coding: utf-8 -*-

"""
Implementation of the Louvain method for community detection.

References
----------
.. [1] Blondel V.D., Guillaume J.-L., Lambiotte R., Lefebvre E. (2008)
   Fast unfolding of communities in large networks. J. Stat. Mech.
   doi:10.1088/1742-5468/2008/10/P10008
"""

import random, itertools
import networkx as nx
from networkx.algorithms.community import modularity


__author__ = 'David Schaller'


class _Supernode:
    """Wrapper class for communities on the previous levels.
    
    The communities found in each level are the nodes in the next level. This
    class contains all nodes of the original graph that belong to such a 
    supernode.
    """
    
    __slots__ = ('nodes',)
    
    def __init__(self):
        
        self.nodes = []
    
    
    def __iter__(self):
        
        return iter(self.nodes)
    
    
    def __next__(self):
        
        pass
    
    
    def append(self, node):
        
        self.nodes.append(node)
    
    
    def extend(self, nodes):
        
        self.nodes.extend(nodes)


class Louvain:
    """
    Implementation of the Louvain method for community detection.
    
    References
    ----------
    .. [1] Blondel V.D., Guillaume J.-L., Lambiotte R., Lefebvre E. (2008)
       Fast unfolding of communities in large networks. J. Stat. Mech.
       doi:10.1088/1742-5468/2008/10/P10008
    """
    
    def __init__(self, graph, weight='weight',
                 at_least_two=False,
                 print_info=False):
        """Constructor for the Louvain class.
        
        Runs the identification of communities.
        
        Parameters
        ----------
        graph : networkx.Graph
            The graph in which communities shall be found.
        weight : str, optional
            The name of the attribute used for edge weighting. The edges of
            `graph` must have this attribute (the default is 'weight', in which
             case edges without this attribute have weight 1.0).
        at_least_two : bool, optional
            If True and the graph has at least two nodes, then the best
            partition is forced to consist of at least two sets, i.e., at least
            two communities are formed (the default is False).
        print_info : bool, optional
            Print the modularity results after each level (the default is
            False).
        """
        
        if not isinstance(graph, nx.Graph) or graph.is_directed():
            raise TypeError("input graph must be an undirected NetworkX graph")
        
        self.orig_graph = graph
        self.weight = weight
        self.at_least_two = at_least_two
        self.print_info = print_info
        
        self._run()
        
    
    def _run(self):
        
        self.partitions = [ [{x} for x in self.orig_graph.nodes()] ]
        
        # modularity is not defined for egdeless / zero-weight graphs
        if not self.orig_graph.size(weight=self.weight):
            self.modularities = [None]
            return
        
        self.modularities = [ modularity(self.orig_graph,
                                         self.partitions[0],
                                         weight=self.weight) ]
        
        # construct the graph for the first level
        graph = self._next_level_graph(self.orig_graph, self.partitions[0])
        
        while True:
            
            level = _Level(graph, weight=self.weight,
                           at_least_two=self.at_least_two)
            
            if not level.moved_on_level:
                break
            
            # partition of the supernodes (= found communities)
            sn_part = list(level.communities.values())
            mod = modularity(graph, sn_part, weight=self.weight)
            
            self.partitions.append(level.get_partition())
            self.modularities.append(mod)
            
            graph = self._next_level_graph(graph, sn_part)
            
            if self.print_info:
                print('----- Level {} -----'.format(len(self.partitions)-1))
                print('computed gain:',
                      self.modularities[-2], '+', level.total_gain,
                      '=', self.modularities[-2] + level.total_gain)
                print('mod. based on supernodes:', mod)
                print('mod. based on original graph',
                      modularity(self.orig_graph, self.partitions[-1],
                                 weight=self.weight))
    
    
    def _next_level_graph(self, graph, partition):
        
        nl_graph = nx.Graph()
        
        # maps the old supernode to the supernode in the next level
        old_to_new = {}
        
        for part_set in partition:
            new_node = _Supernode()
            nl_graph.add_node(new_node)
            for old_node in part_set:
                if isinstance(old_node, _Supernode):
                    new_node.extend(old_node)
                else:
                    new_node.append(old_node)
                old_to_new[old_node] = new_node
        
        for x, y, data in graph.edges(data=True):
            w = data.get(self.weight, 1.0)
            u = old_to_new[x]
            v = old_to_new[y]
            if nl_graph.has_edge(u, v):
                nl_graph[u][v][self.weight] += w
            else:
                nl_graph.add_edge(u, v)
                nl_graph[u][v][self.weight] = w
        
        return nl_graph


class LouvainCustomObj:
    """
    Louvain method for community detection with a customized objetive function.
    
    References
    ----------
    .. [1] Blondel V.D., Guillaume J.-L., Lambiotte R., Lefebvre E. (2008)
       Fast unfolding of communities in large networks. J. Stat. Mech.
       doi:10.1088/1742-5468/2008/10/P10008
    """
    
    def __init__(self, graph,
                 obj_function,
                 minimize=True,
                 args=(),
                 at_least_two=False,
                 print_info=False):
        """Constructor for the Louvain class.
        
        Runs the identification of communities.
        
        Parameters
        ----------
        graph : networkx.Graph
            The graph in which communities shall be found.
        obj_function : function object
            Custom function that is used instead of modularity optimization.
        minimize : bool
            If True, minimize the objetive function, else maximize (the default
            is True, i.e., minimization).
        args : tuple, optional
            Additional arguments relevant for a custom objective function (the
            default is an empty tuple).
        at_least_two : bool, optional
            If True and the graph has at least two nodes, then the best
            partition is forced to consist of at least two sets, i.e., at least
            two communities are formed (the default is False).
        print_info : bool, optional
            Print the modularity results after each level (the default is
            False).
        """
        
        if not isinstance(graph, nx.Graph) or graph.is_directed():
            raise TypeError("input graph must be an undirected NetworkX graph")
        
        self.orig_graph = graph
        self.obj_function = obj_function
        self.minimize = minimize
        self.args = args
        self.at_least_two = at_least_two
        self.print_info = print_info
        
        self._run()
        
    
    def _run(self):
        
        self.partitions = [ [{x} for x in self.orig_graph.nodes()] ]
        self.objectives = [ self.obj_function(self.partitions[0],
                                              *self.args) ]
        
        min_factor = 1 if self.minimize else -1
        
        # construct the graph for the first level
        graph = self._next_level_graph(self.orig_graph, self.partitions[0])
        
        while True:
            
            level = _Level(graph,
                           obj_function=self.obj_function,
                           minimize=self.minimize,
                           args=self.args,
                           at_least_two=self.at_least_two)
            
            if not level.moved_on_level:
                break
            
            # partition of the supernodes (= found communities)
            part = level.get_partition()
            
            self.partitions.append(part)
            self.objectives.append(self.objectives[-1] - 
                                   min_factor * level.total_gain)
            
            graph = self._next_level_graph(graph,
                                           list(level.communities.values()))
            
            if self.print_info:
                print('----- Level {} -----'.format(len(self.partitions)-1))
                print('computed gain:', level.total_gain, '-->',
                      self.objectives[-2] - min_factor * level.total_gain)
                print('mod. based on original graph',
                      self.obj_function(self.partitions[-1],
                                        *self.args))
    
    
    def _next_level_graph(self, graph, partition):
        
        nl_graph = nx.Graph()
        
        # maps the old supernode to the supernode in the next level
        old_to_new = {}
        
        for part_set in partition:
            new_node = _Supernode()
            nl_graph.add_node(new_node)
            for old_node in part_set:
                if isinstance(old_node, _Supernode):
                    new_node.extend(old_node)
                else:
                    new_node.append(old_node)
                old_to_new[old_node] = new_node
        
        for x, y, data in graph.edges(data=True):
            u = old_to_new[x]
            v = old_to_new[y]
            if not nl_graph.has_edge(u, v):
                nl_graph.add_edge(u, v)
        
        return nl_graph
        
        
class _Level:
    """Clustering on a single level in the Louvain method."""
    
    def __init__(self, graph, weight='weight',
                 obj_function=None, minimize=True, args=(),
                 at_least_two=False,):
        
        self.graph = graph
        self.weight = weight
        self.obj_function = obj_function
        self._opt_mode = 1 if minimize else -1
        self.args = args
        self.at_least_two = at_least_two
        
        self.nodes = [x for x in self.graph.nodes()]
        random.shuffle(self.nodes)
        
        self.node_to_com = {x: i for i, x in enumerate(self.nodes)}
        self.communities = {i: {x} for x, i in self.node_to_com.items()}
        
        self.moved_on_level = False
        self.total_gain = 0.0
        
        if self.obj_function is None:
            self._cluster_by_modularity()
        else:
            self._cluster_by_obj()
    
    
    def _cluster_by_modularity(self):
        
        # sum of all edge weight
        m = self.graph.size(weight=self.weight)
        
        # for an edgeless graph, every node is in its own cluster
        if m == 0:
            return
        
        # sum of the weights of the links incident to nodes in the community
        com_tot = {i: 0 for i in self.communities.keys()}
        
        # sum of the weights of all edges incident to nodes
        k = {x: 0 for x in self.nodes}
        
        for x, y, data in self.graph.edges(data=True):
            w = data.get(self.weight, 1.0)
            k[x] += w
            k[y] += w
            com_tot[self.node_to_com[x]] += w
            com_tot[self.node_to_com[y]] += w
    
        while True:
            moved_node = False
            
            for x in self.nodes:
                
                C_x = self.node_to_com[x]
                
                # check if we would merge the last two communities
                if (self.at_least_two and len(self.communities) == 2 and
                    len(self.communities[C_x]) == 1):
                    continue
                
                # maps community C to the sum of weights of the links of x to
                # elements in C \ {x}
                k_x_in = self._weight_sums_to_communities(x)
                
                # remove x from its community
                self.communities[C_x].remove(x)
                com_tot[C_x] -= k[x]
                
                # C_x is preferred in case of ties, i.e. stays in its original
                # community
                
                # equation in Blondel et al. can be simplified to this
                cost_removal = (k_x_in[C_x] - com_tot[C_x] * k[x] / (2 * m)) / m
                best_gain = cost_removal
                best_C = C_x
                visited = {C_x}
                
                for y in self.graph.neighbors(x):
                    
                    C_y = self.node_to_com[y]
                    if C_y in visited:
                        continue
                    
                    new_gain = (k_x_in[C_y] - com_tot[C_y] * k[x] / (2 * m)) / m
                    
                    if new_gain > best_gain:
                        best_gain = new_gain
                        best_C = C_y
                        moved_node = True
                        self.moved_on_level = True
                
                self.communities[best_C].add(x)
                com_tot[best_C] += k[x]
                self.node_to_com[x] = best_C
                self.total_gain += best_gain - cost_removal
                
                # remove the original community if it is now empty
                if len(self.communities[C_x]) == 0:
                    del self.communities[C_x]
            
            # exit the loop when all nodes stayed in their community
            if not moved_node:
                break
    
    
    def _cluster_by_obj(self):
        
        # for an edgeless graph, every node is in its own cluster
        if self.graph.size() == 0:
            return
        
        # avoid recomputation of the partition in each iteration
        part = {i: set(x.nodes) for x, i in self.node_to_com.items()}
        
        obj = self.obj_function(self.get_partition(), *self.args)
    
        while True:
            moved_node = False
            
            for x in self.nodes:
                
                C_x = self.node_to_com[x]
                
                # check if we would merge the last two communities
                if (self.at_least_two and len(self.communities) == 2 and
                    len(self.communities[C_x]) == 1):
                    continue
                
                # remove x from its community
                self.communities[C_x].remove(x)
                part[C_x].difference_update(x.nodes)
                
                # C_x is preferred in case of ties, i.e. stays in its original
                # community
                best_gain = 0.0
                best_C = C_x
                visited = {C_x}
                
                for y in self.graph.neighbors(x):
                    
                    C_y = self.node_to_com[y]
                    if C_y in visited:
                        continue
                    
                    part[C_y].update(x.nodes)
                    new_obj = self.obj_function(part.values(), *self.args)
                    part[C_y].difference_update(x.nodes)
                    
                    new_gain = self._opt_mode * (obj - new_obj)
                    
                    if new_gain > best_gain:
                        best_gain = new_gain
                        best_C = C_y
                        moved_node = True
                        self.moved_on_level = True
                
                self.communities[best_C].add(x)
                part[best_C].update(x.nodes)
                self.node_to_com[x] = best_C
                self.total_gain += best_gain
                obj -= self._opt_mode * best_gain
                
                # remove the original community if it is now empty
                if len(self.communities[C_x]) == 0:
                    del self.communities[C_x]
                    del part[C_x]
            
            # exit the loop when all nodes stayed in their community
            if not moved_node:
                break
            
    
    def get_partition(self):
        """Convert the communities of supernodes into a partition of the 
        orginal nodes."""
        
        return [set(itertools.chain.from_iterable(com))
                for com in self.communities.values()]
    
    
    def _weight_sums_to_communities(self, x):
    
        # sum of the weight from x to every community C ( \{x} )
        # the community of x must be present even if x is its only element
        weight_sums = {self.node_to_com[x]: 0.0}
        
        for y in self.graph.neighbors(x):
            if x == y:
                continue
            C = self.node_to_com[y]
            w = self.graph.get_edge_data(x, y).get(self.weight, 1.0)
            weight_sums[C] = weight_sums.get(C, 0.0) + w
        
        return weight_sums


if __name__ == '__main__':
    
    # random graph
    G = nx.erdos_renyi_graph(100, 0.02)
    louv = Louvain(G, print_info=False)
    for mod, part in zip(louv.modularities, louv.partitions):
        print(mod)
    
    # example from Blondel et al. 2008
    G2 = nx.Graph()
    G2.add_nodes_from([x for x in range(16)])
    G2.add_edges_from([(0,2), (0,3), (0,4), (0,5),
                       (1,2), (1,4), (1,7), (2,4), (2,5), (2,6),
                       (3,7), (4, 10), (5,7), (5,11), (6,7), (6,11),
                       (8,9), (8,10), (8,11), (8,14), (8,15), (9,12), (9,14),
                       (10,11), (10,12), (10,13), (10,14), (11,13),
                       ])
    
    louv = Louvain(G2, print_info=True)
    for mod, part in zip(louv.modularities, louv.partitions):
        print(part, mod)



# -*- coding: utf-8 -*-

import itertools

import gurobipy as gp
from gurobipy import GRB

import networkx as nx
import tralda.tools.GraphTools as gt


__author__ = 'David Schaller'


class EditorSuperClass:
    
    status = {1: 'LOADED',
              2: 'OPTIMAL',
              3: 'INFEASIBLE',
              4: 'INF_OR_UNBD',
              5: 'UNBOUNDED',
              6: 'CUTOFF',
              7: 'ITERATION_LIMIT',
              8: 'NODE_LIMIT',
              9: 'TIME_LIMIT',
              10: 'SOLUTION_LIMIT',
              11: 'INTERRUPTED',
              12: 'NUMERIC',
              13: 'SUBOPTIMAL',
              14: 'INPROGRESS',
              15: 'USER_OBJ_LIMIT',}
    
    
    def __init__(self, graph):
        
        self.graph = graph
        self.n = self.graph.order()
        self.color_dict = gt.sort_by_colors(graph)
        
    
    def build_model(self):
        
        # constant values repr. the edges in the original graph
        self.E = {(x, y): int(self.graph.has_edge(x, y))
                     for x in self.graph.nodes()
                     for y in self.graph.nodes()
                     if x != y}
        
        # build model with variables and objective function
        self.model = gp.Model()
        
        # binary varibles pairs of vertices
        self.e = self.model.addVars(self.E, vtype=GRB.BINARY, name='e')
        
        # finally sink all variables
        self.model.update()
        
        # set objective
        self._objective()
        
    
    def _objective(self):
        """Sets the objective function."""
        
        self.obj = gp.LinExpr()
        
        # add (1 - e_xy) * E_xy + (1 - E_xy) * e_xy for all x, y
        for x, y in self.E:
            if self.E[x, y]:
                self.obj += 1 - self.e[x, y]
            else:
                self.obj += self.e[x, y]
        
        self.model.setObjective(self.obj, GRB.MINIMIZE)
                                
                                    
    def optimize(self, time_limit=False):
        """Solves the editing problem."""
        
        # set new time limit temporarily
        if time_limit:
            original_time_limit = self.model.Params.timeLimit
            self.model.Params.timeLimit = time_limit
        
        self.model.optimize()
        
        # reset time limit
        if time_limit:
            self.model.Params.timeLimit = original_time_limit
    
    
    def _sink_free_colored(self):
        """Ensures that each vertex has out-arcs to every other color."""
        
        for x in self.graph.nodes():
            for col in self.color_dict:
                if self.graph.nodes[x]['color'] != col:
                    self.model.addConstr(
                        gp.quicksum((self.e[x, y]
                                     for y in self.color_dict[col])) >= 1,
                        name='c[{},{}]'.format(x, col)
                        )
        
        
    def get_solution(self):
        """Objective value and resulting digraph of the (current) solution."""
        
        # if self.model.status != GRB.OPTIMAL:
        #     raise RuntimeError('could not find optimal solution')
        
        sol_graph = nx.DiGraph()
        sol_graph.add_nodes_from(self.graph.nodes.data())
        
        for edge, value in self.model.getAttr('x', self.e).items():
            if value > 0.5:     # handle potential numerical inaccuracies
                sol_graph.add_edge(*edge)
        
        
        return self.model.ObjVal, sol_graph
    
    
    def get_status(self):
        
        return BMGEditor.status[self.model.status]
    
    
    def get_solve_time(self):
        
        return self.model.Runtime
    
    


class BMGEditor(EditorSuperClass):
    
    
    def __init__(self, graph):
        
        super().__init__(graph)
        
    
    def build_model(self):
        
        # arc variables and objective
        super().build_model()
        
        self._additional_variables()
        
        # subject to constraints
        self._sink_free_colored()
        self._inf_forb_triples()
        self._proper_hierarchy()
        
    
    def _additional_variables(self):
        
        # binary variable for each triple ab|c (= ba|c)
        triples = ((a, b, c)
                   for a, b, c in itertools.permutations(self.graph.nodes(), 3)
                   if a < b)
        self.t = self.model.addVars(triples, vtype=GRB.BINARY, name='t')
        
        # matrix for hierarchy repr. the tree
        self.M = self.model.addVars(((a, p) for a in self.graph.nodes()
                                            for p in range(self.n-2)),
                                vtype=GRB.BINARY, name='M')
        
        # a, b in custer p, but not c ?
        self.m = self.model.addVars(((a, b, c, p) for a, b, c in self.t
                                                  for p in range(self.n-2)),
                                    vtype=GRB.BINARY, name='m')
        
        # varibles for three-gamete condition
        self.C = self.model.addVars(
            ((p, q, gam) for p, q in itertools.combinations(range(self.n-2), 2)
                         for gam in range(3)),
            vtype=GRB.BINARY, name='C')
        
        # finally sink all variables
        self.model.update()
        
#        for v in self.model.getVars():
#            print(v)
    
    
    def _inf_forb_triples(self):
        """Ensures that all informative triples of the graph are displayed
        by the tree but none of the forbidden triples."""
        
        for x in self.graph.nodes():
            for col in self.color_dict:
                if self.graph.nodes[x]['color'] != col:
                    for y1, y2 in itertools.permutations(self.color_dict[col],
                                                         2):
                        if x < y1:
                            # informative triple x y1 | y2
                            self.model.addConstr(
                                (self.e[x, y1] - self.e[x, y2] -
                                 self.t[x, y1, y2] <= 0),
                                name='it[{},{},{}]'.format(x, y1, y2)
                                )
                            # forbidden triple x y1 | y2
                            self.model.addConstr(
                                (self.e[x, y1] + self.e[x, y2] +
                                 self.t[x, y1, y2] <= 2),
                                name='ft[{},{},{}]'.format(x, y1, y2)
                                )
                        else:
                            # informative triple x y1 | y2
                            self.model.addConstr(
                                (self.e[x, y1] - self.e[x, y2] -
                                 self.t[y1, x, y2] <= 0),
                                name='it[{},{},{}]'.format(x, y1, y2)
                                )
                            # forbidden triple x y1 | y2
                            self.model.addConstr(
                                (self.e[x, y1] + self.e[x, y2] +
                                 self.t[y1, x, y2] <= 2),
                                name='ft[{},{},{}]'.format(x, y1, y2)
                                )
        
        # set m((ab|c), p) correctly
        for a, b, c, p in self.m:
            
            expr = (self.M[a, p] + self.M[b, p] + (1 - self.M[c, p]) - 
                    3 * self.m[a, b, c, p])
            self.model.addConstr(expr >= 0)
            self.model.addConstr(expr <= 2)
        
        # informative and forbidden triples
        for a, b, c in self.t:
            expr = gp.quicksum((self.m[a, b, c, p] for p in range(self.n-2)))
            self.model.addConstr(self.t[a, b, c] <= expr)
            self.model.addConstr(expr <= (self.n - 2) * self.t[a, b, c])
        
        
    def _proper_hierarchy(self):
        """Ensures that a proper tree is constructed by the three-gamete
        condition."""
        
        self.model.addConstrs(
            (self.C[p, q, 0] >= self.M[a, q] - self.M[a, p]
             for p, q in itertools.combinations(range(self.n-2), 2)
             for a in self.graph.nodes()))
        self.model.addConstrs(
            (self.C[p, q, 1] >= self.M[a, p] - self.M[a, q]
             for p, q in itertools.combinations(range(self.n-2), 2)
             for a in self.graph.nodes()))
        self.model.addConstrs(
            (self.C[p, q, 2] >= self.M[a, q] + self.M[a, p] - 1
             for p, q in itertools.combinations(range(self.n-2), 2)
             for a in self.graph.nodes()))
                               
        self.model.addConstrs(
            (self.C[p, q, 0] + self.C[p, q, 1] + self.C[p, q, 2] <= 2
             for p, q in itertools.combinations(range(self.n-2), 2)))
    
    
class TwoBMGEditor(EditorSuperClass):

    def __init__(self, graph):

        super().__init__(graph)
        
        if len(self.color_dict) != 2:
            raise RuntimeError('not a 2-colored digraph')


    def build_model(self):
        
        # arc variables and objective
        super().build_model()
        
        # subject to constraints
        self._sink_free_colored()
        self._forbidden_subgraphs()


    def _forbidden_subgraphs(self):
        """Ensures that the graph does not contain a (2-colored) forbidden
        induced subgraph."""

        for col1, col2 in itertools.permutations(self.color_dict, 2):
            for x1, x2 in itertools.permutations(self.color_dict[col1], 2):

                for y1, y2 in itertools.permutations(self.color_dict[col2], 2):

                    # type (F1)
                    self.model.addConstr(
                        (self.e[x1, y1] + self.e[y1, x2] + self.e[y2, x2] - 
                         self.e[x1, y2] - self.e[y2, x1] <= 2),
                        name='F1[{},{},{},{}]'.format(x1, x2, y1, y2)
                        )

                    # type (F2)
                    self.model.addConstr(
                        (self.e[x1, y1] + self.e[y1, x2] + self.e[x2, y2] - 
                         self.e[x1, y2] <= 2),
                        name='F2[{},{},{},{}]'.format(x1, x2, y1, y2)
                        )

                for y1, y2, y3 in itertools.permutations(self.color_dict[col2],
                                                         3):
                    # since type (F3) subgraphs are symmetrical
                    if y1 > y2:
                        continue

                    # type (F3)
                    self.model.addConstr(
                        (self.e[x1, y1] + self.e[x1, y3] +
                         self.e[x2, y2] + self.e[x2, y3] - 
                         self.e[x1, y2] - self.e[x2, y1] <= 3),
                        name='F3[{},{},{},{},{}]'.format(x1, x2, y1, y2, y3)
                        )


class BinaryBMGEditor(EditorSuperClass):
    
    
    def __init__(self, graph):
        
        super().__init__(graph)
        
    
    def build_model(self):
        
        # arc variables and objective
        super().build_model()
        
        self._additional_variables()
        
        # subject to constraints
        self._sink_free_colored()
        self._ext_inf_triples()
        self._triple_consistency()
        
    
    def _additional_variables(self):
        
        # binary variable for each triple ab|c (= ba|c)
        triples = ((a, b, c)
                   for a, b, c in itertools.permutations(self.graph.nodes(), 3)
                   if a < b)
        self.t = self.model.addVars(triples, vtype=GRB.BINARY, name='t')
        
        # finally sink all variables
        self.model.update()
    
    
    def _ext_inf_triples(self):
        """Ensures that all informative triples of the graph are displayed
        by the tree but none of the forbidden triples."""
        
        for x in self.graph.nodes():
            for col in self.color_dict:
                if self.graph.nodes[x]['color'] != col:
                    for y1, y2 in itertools.permutations(self.color_dict[col],
                                                         2):
                        if x < y1:
                            # informative triple x y1 | y2
                            self.model.addConstr(
                                (self.e[x, y1] - self.e[x, y2] -
                                 self.t[x, y1, y2] <= 0),
                                name='it[{},{},{}]'.format(x, y1, y2)
                                )
                        else:
                            # informative triple x y1 | y2
                            self.model.addConstr(
                                (self.e[x, y1] - self.e[x, y2] -
                                 self.t[y1, x, y2] <= 0),
                                name='it[{},{},{}]'.format(x, y1, y2)
                                )
                        
                        # one direction is sufficient
                        if y1 < y2:
                            self.model.addConstr(
                                (self.e[x, y1] + self.e[x, y2] -
                                 self.t[y1, y2, x] <= 1),
                                name='it[{},{},{}]'.format(x, y1, y2)
                                )
    
    
    def _triple_consistency(self):
        """Ensures concistency of all triples."""
        
        # strictly dense triple set
        for triple in itertools.combinations(self.graph.nodes(), 3):
            a, b, c = sorted(triple)
            
            self.model.addConstr((self.t[a, b, c] + self.t[a, c, b] +
                                  self.t[b, c, a] == 1),
                                 name='d[{},{},{}]'.format(a, b, c))
                                    
        # compatibility based on 2-order inference rules
        for a, b, c, d in itertools.permutations(self.graph.nodes(), 4):
            #a, b, c, d = sorted(triple)
            
            ab = sorted([a, b])
            ad = sorted([a, d])
            bd = sorted([b, d])
                    
            self.model.addConstr((2 * self.t[(*ab, c)] + 2 * self.t[(*ad, b)] -
                                  self.t[(*bd, c)] - self.t[(*ad, c)] <= 2),
                                 name='cp[{},{},{},{}]'.format(a, b, c, d))
        

if __name__ == '__main__':
    
    from asymmetree.tools.PhyloTreeTools import random_colored_tree
    from asymmetree.analysis import (bmg_from_tree,
                                     is_bmg,
                                     binary_refinable_tree)
    
    # i = 0
    # while True:
    #     i += 1
    
    random_tree = random_colored_tree(12, 4,
                                      # binary=True,
                                      force_all_colors=True,)
    bmg = bmg_from_tree(random_tree)
    graph = gt.disturb_graph(bmg, 0.2, 0.2)
    
    # solver = BMGEditor(graph)
    solver = BinaryBMGEditor(graph)
    solver.build_model()
    solver.optimize(time_limit=None)
    opt_val, sol_graph = solver.get_solution()
    
    print('Status:', solver.get_status())
    print('Runtime:', solver.get_solve_time(), 's')
    
    print('-------------')
    print([(x, bmg.nodes[x]['color']) for x in bmg.nodes])
    print(bmg.edges)
    print('-------------')
    print([(x, sol_graph.nodes[x]['color']) for x in sol_graph.nodes])
    print(sol_graph.edges)
    print('-------------')
    
    print('original vs disturbed:', gt.symmetric_diff(bmg, graph))
    print('original vs new:', gt.symmetric_diff(sol_graph, bmg))
    print('disturbed vs new', gt.symmetric_diff(sol_graph, graph),
          'should be', opt_val,
          'and <=', gt.symmetric_diff(bmg, graph))
    print(sol_graph.order(), sol_graph.size())
    print('is BMG orig:', True if is_bmg(bmg) else False)
    print('is BMG dist:', True if is_bmg(graph) else False)
    success = True if is_bmg(sol_graph) else False
    print('is BMG new:', success)
    be_success = True if binary_refinable_tree(sol_graph, mincut=False) else False
    print('is beBMG new:', be_success)
    
    colors = ['lightblue', 'red', 'green', 'yellow', 'black']
    color_map = []
    for v in sol_graph:
        color_map.append(colors[sol_graph.nodes[v]['color']-1])
    nx.draw(sol_graph,
            node_color=color_map,
            with_labels=True)
    
        # print('iteration', i)
        # if not success:
        #     break
    


# -*- coding: utf-8 -*-

"""
Algorithms for graph (bi)partitioning.
"""

import random, math
import networkx as nx

from tralda.datastructures import LinkedList, TreeSet


__author__ = 'David Schaller'


# ----------------------------------------------------------------------------
#                           Karger's algorithm
#
# Implementation of Karger's algorithm for minimal edge cut. Randomized
# algorithm that returns a minimal edge cut with high probability.
# This implementation is based on: 
# http://web.stanford.edu/class/archive/cs/cs161/cs161.1166/lectures/
# lecture15.pdf
# ----------------------------------------------------------------------------


class _Supernode:
    
    def __init__(self, initial_node):
        
        self.nodes = LinkedList()
        self.nodes.append(initial_node)
        
    
    def __len__(self):
        
        return len(self.nodes)
    
    
    def __lt__(self, other):
        
        return id(self) < id(other)
    
    
    def merge_from(self, other):
        
        self.nodes.concatenate(other.nodes)
        

def _sort2(a, b):
    
    return (a, b) if a < b else (b, a)
        
        
class Karger():
    
    def __init__(self, graph):
        
        if not isinstance(graph, nx.Graph):
            raise TypeError('must be a NetworkX graph')
            
        if not graph.order() >= 2:
            raise ValueError('graph must have >= 2 nodes')
            
        if not nx.is_connected(graph):
            raise ValueError('graph not connected')
            
        self.graph = graph
            
    
    def run(self):
        """Execute a single run of the contraction procedure."""
        
        return self._karger_main()
    
    
    def best_out_of(self, runs=None):
        """Execute multiple runs and return the one with the smallest cutvalue.
        
        Keyword arguments:
            runs -- number of runs; default is None, in which case
                O(n^2 * log n) runs are executed
        """
        
        if not runs:
            runs = self._default_run_number()
        
        best_V1, best_V2, best_cutvalue = None, None, None
        
        for _ in range(runs):
            V1, V2, cutvalue = self._karger_main()
            
            if not best_cutvalue or cutvalue < best_cutvalue:
                best_V1, best_V2, best_cutvalue = V1, V2, cutvalue
                
        return best_V1, best_V2, best_cutvalue
    
    
    def generate(self, runs=None):
        """Generator for runs of the contraction procedure.
        
        Keyword arguments:
            runs -- number of runs; default is None, in which case
                O(n^2 * log n) runs are generated
        """
        
        if not runs:
            runs = self._default_run_number()
        
        for _ in range(runs):
            yield self._karger_main()
            
            
    def _default_run_number(self):
        
        n = self.graph.order()
        return round(n**2 * math.log(n))
    
        
    def _full_edge_set(self):
        
        # construct the balanced search tree for the full edge set once
        if not hasattr(self, 'full_egde_set'):
            self.full_egde_set = TreeSet()
            for u, v in self.graph.edges():
                self.full_egde_set.add( (u, v) )
        
        # otherwise copy which is O(|E|)
        return self.full_egde_set.copy()
        
        
    def _initialize(self):
        
        self.supernodes = set()
        self.supernode_dict = {}            # maps node --> supernode
        self.superedges = {}                # maps superedge --> edges
        self.edges = self._full_edge_set()  # set of all (non-loop) edges
        
        for v in self.graph.nodes():
            sn = _Supernode(v)
            self.supernode_dict[v] = sn
            self.supernodes.add(sn)
        
        for u, v in self.graph.edges():
            u_sn = self.supernode_dict[u]
            v_sn = self.supernode_dict[v]
            e = _sort2(u_sn, v_sn)
            self.superedges[e] = LinkedList( [(u, v)] )
            
            
    def _merge(self, a, b):
        """Merge two supernodes a and b."""
        
        # bigger supernode becomes the new supernode x
        if len(a) >= len(b):
            x = a
            other = b
        else:
            x = b
            other = a
            
        for v in other.nodes:
            self.supernode_dict[v] = x
        x.merge_from(other)
        self.supernodes.remove(other)
        
        for d in self.supernodes:
            
            if d is x:
                continue
            
            # merge the edge lists of superedges ad and bd
            edges_xd = LinkedList()
            for se in (_sort2(a, d), _sort2(b, d)):
                if se in self.superedges:
                    edges_xd.concatenate(self.superedges[se])
                    del self.superedges[se]
            if edges_xd:
                self.superedges[_sort2(x, d)] = edges_xd
                
                
    def _karger_main(self):
        
        self._initialize()
        
        while len(self.supernodes) > 2:
            
            u, v = random.choice(self.edges)
            u_sn = self.supernode_dict[u]
            v_sn = self.supernode_dict[v]
            
            self._merge(u_sn, v_sn)
            
            se = _sort2(u_sn, v_sn)
            self.edges.difference_update(self.superedges[se])
            del self.superedges[se]
        
        x = self.supernodes.pop()
        y = self.supernodes.pop()

        cut_value = len(self.superedges[_sort2(x, y)])
        # print('cut_edges', list(self.superedges[_sort2(x, y)]))
            
        return cut_value, [list(x.nodes), list(y.nodes)]
        
    

if __name__ == '__main__':
    
    G = nx.Graph()
    G.add_edges_from([('a', 'b'), ('a', 'c'), ('a', 'd'),
                      ('b', 'd'), ('c', 'd'), ('c', 'e'), ('d', 'e')])
    
    karger = Karger(G)
    for result in karger.generate(3):
        print(*result)
       



# -*- coding: utf-8 -*-

"""
File I/O related to module 'Scenario'.
"""

import itertools

import networkx as nx

from asymmetree.tools.PhyloTreeTools import to_newick


__author__ = 'David Schaller'


def matrix_to_phylip(filename, leaves, matrix):
    """Write the distance matrix in phylip format."""
    
    with open(filename, 'w') as f:
        f.write(str(len(leaves)))
        for i in range(len(leaves)):
            f.write('\n' + str(leaves[i].label))
            for j in range(len(leaves)):
                f.write('\t' + str(matrix[i,j]))


def species_to_genes(filename, scenario):
    """Write the species and corresponding genes file."""
    
    with open(filename, 'w') as f:
        species = list(scenario.reconc_dict)
        for i in range(len(species)):
            f.write(str(species[i]))
            for gene in scenario.reconc_dict[species[i]]:
                f.write('\t' + str(gene.label))
            if i < len(species)-1:
                f.write('\n')


def species_pairs_outgroups(filename, scenario):
    """Write pairs of species with corresponding outgroups 
    
    ("ok-species" file)."""
    
    with open(filename, 'w') as f:
        begin = True
        for i in range(len(scenario.subtree_list)):
            for speciesX, speciesY in itertools.combinations(scenario.subtree_list[i], 2):
                if begin:
                    f.write('{}\t{}'.format(speciesX, speciesY))
                    begin = False
                else:
                    f.write('\n{}\t{}'.format(speciesX, speciesY))
                for j in range(len(scenario.subtree_list)):
                    if i == j:
                        continue
                    for outgroup_species in scenario.subtree_list[j]:
                        f.write('\t' + str(outgroup_species))
   
                 
def species_in_subtree(filename, scenario):
    """Species in the same subtree under the the root
    
    (one line per subtree)."""
    
    with open(filename, 'w') as f:
        begin = True
        for sp_list in scenario.subtree_list:
            if begin:
                f.write('\t'.join([str(item) for item in sp_list]))
                begin = False
            else:
                f.write('\n')
                f.write('\t'.join([str(item) for item in sp_list]))


def subtree_bmg(filename, scenario):
    """Write the (subgraph of) BMG."""
    
    with open(filename, 'w') as f:
        begin = True
        for u, v in scenario.bmg_subtrees.edges:
            if begin:
                f.write('{}\t{}'.format(u, v))
                begin = False
            else:
                f.write('\n{}\t{}'.format(u, v))


def parse_bmg_edges(output, scenario):
    """Parse edge output from external program to a graph
    
    (converts keys to ints if possible)."""
    
    bmg = nx.DiGraph()
    bmg.add_nodes_from(scenario.bmg.nodes(data=True))
    
    for line in output.split("\n"):
        if line:
            edge = line.split()
            if len(edge) >= 2:
                u = int(edge[0]) if edge[0].isdigit() else edge[0]
                v = int(edge[1]) if edge[1].isdigit() else edge[1]
                if v not in bmg.nodes:
                    print('Not in BMG', v)
                if u not in bmg.nodes:
                    print('Not in BMG', u)
                bmg.add_edge(u, v)
                
    return bmg


def write_newick(filename, tree):
    """Write Newick tree."""
    
    with open(filename, 'w') as f:
        f.write(to_newick(tree))



# -*- coding: utf-8 -*-

"""
Extended Best Hits Method.

Implementation of the Extended Best Hits method for best match inference.
"""

import os, subprocess, time

from asymmetree.analysis.BestMatches import extended_best_hits
from asymmetree.file_io.ScenarioFileIO import (parse_bmg_edges,
                                               matrix_to_phylip,
                                               species_to_genes)


__author__ = 'David Schaller'


# --------------------------------------------------------------------------
#                           PYTHON IMPLEMENTATION
#
# --------------------------------------------------------------------------
    
def ebh(leaves, D, epsilon=1e-8):
    """Compute BMG and RBMG from a distances matrix D.
    
    Keyword arguments:
        epsilon -- epsilon for relative BM threshold: (x,y) in BMG if
                   D(x,y) <= (1+epsilon) * min d(x,y'),
                   default=1e-8 (for limited float precision).
    """
    
    return extended_best_hits(leaves, D, epsilon=epsilon, supply_rbmg=True)


# --------------------------------------------------------------------------
#                            EXTERNAL C++ PROGRAM
#
# --------------------------------------------------------------------------
 
def ebh_qinfer(scenario,
               matrix_filename, species_filename,
               epsilon=0.5,
               benchmark_file=None,
               binary_path=None):
    """Compute BMG and RBMG from a distances matrix D using 'qinfer'.
    
    Keyword arguments:
        epsilon -- epsilon for relative BM threshold: (x,y) in BMG if
                   D(x,y) <= (1+epsilon) * min d(x,y'),
                   default=10E-8 (for limited float precision).
        benchmark_file -- activate benchmarking in 'qinfer' and
                          specify the filename
        binary_path -- path to 'qinfer' binary (if not available
                       within path)
    """
    
    if not binary_path:
        qinfer_command = "qinfer"
    elif os.path.exists(binary_path):
        qinfer_command = binary_path
    else:
        raise FileNotFoundError("path to qinfer binary file '{}' does not exist".format(binary_path))
    
    output = -1
    command = [qinfer_command, matrix_filename, species_filename,
               "--disable-quartet", "--epsilon=" + str(epsilon)]

    if benchmark_file is not None:
        command.append( "--benchmark=" + benchmark_file )
    
    # call 'qinfer' and measure execution time
    start = time.time()
    
    try:
        output = subprocess.run(command, stdout=subprocess.PIPE)
    except:
        raise FileNotFoundError("calling qinfer failed")
    
    exec_time = time.time() - start
    
    if output == -1:
        raise Exception("no output from qinfer")
    
    bmg = parse_bmg_edges(output.stdout.decode(), scenario)
    
    return bmg, bmg.to_undirected(reciprocal=True), exec_time


def ebh_from_scenario(scenario, epsilon=0.5):
    """Compute BMG and RBMG from a scenario using 'qinfer'.
    
    Keyword arguments:
        epsilon -- epsilon for relative BM threshold: (x,y) in BMG if
                   D(x,y) <= (1+epsilon) * min d(x,y'),
                   default=10E-8 (for limited float precision).
    """

    matrix_filename = "temp.phylip"
    species_filename = "temp_species.txt"
    
    matrix = scenario.get_distance_matrix()
    matrix_to_phylip(matrix_filename, scenario.genes, matrix)
    species_to_genes(species_filename, scenario)
    
    bmg, rbmg, exec_time = ebh_qinfer(scenario,
                                      matrix_filename, species_filename,
                                      epsilon=epsilon)
    
    os.remove(matrix_filename)
    os.remove(species_filename)
    
    return bmg, rbmg, exec_time


# -*- coding: utf-8 -*-

import itertools

from docplex.mp.model import Model

import networkx as nx
import tralda.tools.GraphTools as gt


__author__ = 'David Schaller'


class BMGEditor:
    
    
    def __init__(self, graph):
        
        self.graph = graph
        self.n = self.graph.order()
        self.color_dict = gt.sort_by_colors(graph)
        
    
    def build_model(self):
        
        # constant values repr. the edges in the original graph
        self.E = {(x, y): int(self.graph.has_edge(x, y))
                     for x in self.graph.nodes()
                     for y in self.graph.nodes()
                     if x != y}
        
        # build model with variables and objective function
        self.model = Model('BMG_Editing')
        self._variables()
        
        # subject to constraints
        self._well_colored()
        self._inf_forb_triples()
        self._proper_hierarchy()
        
        self._objective()
        
        self.model.print_information()
        
    
    def _variables(self):
        
        # binary varibles pairs of vertices
        self.e = self.model.binary_var_dict(self.E, name='e')
        
        # binary variable for each triple ab|c (= ba|c)
        triples = ((a, b, c)
                   for a, b, c in itertools.permutations(self.graph.nodes(), 3)
                   if a < b)
        self.t = self.model.binary_var_dict(triples, name='t')
        
        # matrix for hierarchy repr. the tree
        self.M = self.model.binary_var_matrix(self.graph.nodes(),
                                              range(self.n-2),
                                              name='M')
        
        # a, b in custer p, but not c ?
        self.m = self.model.binary_var_dict(((a, b, c, p)
                                             for a, b, c in self.t
                                             for p in range(self.n-2)),
                                            name='m')
        
        # varibles for three-gamete condition
        self.C = self.model.binary_var_dict(
            ((p, q, gam) for p, q in itertools.combinations(range(self.n-2), 2)
                         for gam in range(3)),
             name='C')
        
    
    def _objective(self):
        """Sets the objective function."""
        
        self.obj = self.model.linear_expr()
        
        # add (1 - e_xy) * E_xy + (1 - E_xy) * e_xy for all x, y
        for x, y in self.E:
            if self.E[x, y]:
                self.obj += 1 - self.e[x, y]
            else:
                self.obj += self.e[x, y]
        
        self.model.minimize(self.obj)
        
    
    def _well_colored(self):
        """Ensures that each vertex has out-arcs to every other color."""
        
        for x in self.graph.nodes():
            for col in self.color_dict:
                if self.graph.nodes[x]['color'] != col:
                    self.model.add_constraint(
                        self.model.sum((self.e[x, y]
                                       for y in self.color_dict[col])) >= 1,
                        ctname='c[{},{}]'.format(x, col)
                        )
    
    
    def _inf_forb_triples(self):
        """Ensures that all informative triples of the graph are displayed
        by the tree but none of the forbidden triples."""
        
        for x in self.graph.nodes():
            for col in self.color_dict:
                if self.graph.nodes[x]['color'] != col:
                    for y1, y2 in itertools.permutations(self.color_dict[col],
                                                         2):
                        if x < y1:
                            # informative triple x y1 | y2
                            self.model.add_constraint(
                                (self.e[x, y1] - self.e[x, y2] -
                                 self.t[x, y1, y2] <= 0),
                                ctname='it[{},{},{}]'.format(x, y1, y2)
                                )
                            # forbidden triple x y1 | y2
                            self.model.add_constraint(
                                (self.e[x, y1] + self.e[x, y2] +
                                 self.t[x, y1, y2] <= 2),
                                ctname='ft[{},{},{}]'.format(x, y1, y2)
                                )
                        else:
                            # informative triple x y1 | y2
                            self.model.add_constraint(
                                (self.e[x, y1] - self.e[x, y2] -
                                 self.t[y1, x, y2] <= 0),
                                ctname='it[{},{},{}]'.format(x, y1, y2)
                                )
                            # forbidden triple x y1 | y2
                            self.model.add_constraint(
                                (self.e[x, y1] + self.e[x, y2] +
                                 self.t[y1, x, y2] <= 2),
                                ctname='ft[{},{},{}]'.format(x, y1, y2)
                                )
        
        # set m((ab|c), p) correctly
        for a, b, c, p in self.m:
            
            expr = (self.M[a, p] + self.M[b, p] + (1 - self.M[c, p]) - 
                    3 * self.m[a, b, c, p])
            self.model.add_constraint(expr >= 0)
            self.model.add_constraint(expr <= 2)
        
        # informative and forbidden triples
        for a, b, c in self.t:
            expr = self.model.sum((self.m[a, b, c, p] for p in range(self.n-2)))
            self.model.add_constraint(self.t[a, b, c] <= expr)
            self.model.add_constraint(expr <= (self.n - 2) * self.t[a, b, c])
        
        
    def _proper_hierarchy(self):
        """Ensures that a proper tree is constructed by the three-gamete
        condition."""
        
        self.model.add_constraints(
            (self.C[p, q, 0] >= self.M[a, q] - self.M[a, p]
             for p, q in itertools.combinations(range(self.n-2), 2)
             for a in self.graph.nodes()))
        self.model.add_constraints(
            (self.C[p, q, 1] >= self.M[a, p] - self.M[a, q]
             for p, q in itertools.combinations(range(self.n-2), 2)
             for a in self.graph.nodes()))
        self.model.add_constraints(
            (self.C[p, q, 2] >= self.M[a, q] + self.M[a, p] - 1
             for p, q in itertools.combinations(range(self.n-2), 2)
             for a in self.graph.nodes()))
                               
        self.model.add_constraints(
            (self.C[p, q, 0] + self.C[p, q, 1] + self.C[p, q, 2] <= 2
             for p, q in itertools.combinations(range(self.n-2), 2)))        
                                
                                    
    def optimize(self, time_limit=False):
        """Solves the editing problem."""
        
        # set new time limit temporarily
        if time_limit:
            original_time_limit = self.model.time_limit
            self.model.set_time_limit(time_limit)
        
        self.solution = self.model.solve(url=None, key=None)
        
        # reset time limit
        if time_limit:
            self.model.time_limit = original_time_limit
        
        print('Solve details:')
        print(self.model.get_solve_details())
        
        
    def get_solution(self):
        """Objective value and resulting digraph of the (current) solution."""
        
        sol_graph = nx.DiGraph()
        sol_graph.add_nodes_from(self.graph.nodes.data())
        
        for edge, variable in self.e.items():
            
            # handle potential numerical inaccuracies
            if self.solution.get_value(variable) > 0.5:
                sol_graph.add_edge(*edge)
        
        
        return self.solution.get_objective_value(), sol_graph
    
    
    def get_status(self):
        
        return self.solution.solve_details.status
    
    
    def get_solve_time(self):
        
        return self.solution.solve_details.time
        

if __name__ == '__main__':
    
    from asymmetree.tools.PhyloTreeTools import random_colored_tree
    from asymmetree.analysis import (bmg_from_tree,
                                     is_bmg,)
    
    random_tree = random_colored_tree(7, 4, force_all_colors=True)
    bmg = bmg_from_tree(random_tree)
    graph = gt.disturb_graph(bmg, 0.2, 0.2)
    
    solver = BMGEditor(graph)
    solver.build_model()
    solver.optimize(time_limit=3)
    solver.get_solution()
    opt_val, sol_graph = solver.get_solution()
    
    print('Status:', solver.get_status())
    print('Runtime:', solver.get_solve_time(), 's')
    
    print('-------------')
    print([(x, bmg.nodes[x]['color']) for x in bmg.nodes])
    print(bmg.edges)
    print('-------------')
    print([(x, sol_graph.nodes[x]['color']) for x in sol_graph.nodes])
    print(sol_graph.edges)
    print('-------------')
    
    print('original vs disturbed:', gt.symmetric_diff(bmg, graph))
    print('original vs new:', gt.symmetric_diff(sol_graph, bmg))
    print('disturbed vs new', gt.symmetric_diff(sol_graph, graph),
          'should be', opt_val,
          'and <=', gt.symmetric_diff(bmg, graph))
    print(sol_graph.order(), sol_graph.size())
    print('is BMG orig:', True if is_bmg(bmg) else False)
    print('is BMG dist:', True if is_bmg(graph) else False)
    success = True if is_bmg(sol_graph) else False
    print('is BMG new:', success)
    


# -*- coding: utf-8 -*-

import random, itertools
import numpy as np


__author__ = 'David Schaller'


# ----------------------------------------------------------------------------
#                           Greedy bipartition
# ----------------------------------------------------------------------------

def partition_cut_value(partition, G):
    
    cut_value = 0
    
    for Vi, Vj in itertools.combinations(partition, 2):
        for x, y in itertools.product(Vi, Vj):
            if G.has_edge(x, y):
                cut_value += 1
    
    return cut_value
    

def _move(from_set, to_set, x):
    
    from_set.remove(x)
    to_set.add(x)


def greedy_bipartition(V, f_obj, minimize=True, args=()):
    """Randomized greedy bipartitioning with custom objective function."""
    
    if len(V) < 2:
        raise ValueError('iterable must have >=2 elements')
    
    V1, V2 = set(), set(V)
    remaining = set(V)
    
    best_obj = float('inf') if minimize else float('-inf')
    best_bp = None
    
    for _ in range(len(V)-1):
        
        best_obj_local = float('inf') if minimize else float('-inf')
        best_x = None
        
        for x in remaining:
            _move(V2, V1, x)
            obj = f_obj([V1, V2], *args)
            if ((minimize and obj < best_obj_local) or
                (not minimize and obj > best_obj_local)):
                best_obj_local = obj
                best_x = [x]
            elif obj == best_obj_local:
                best_x.append(x) 
            _move(V1, V2, x)
            
        x = random.choice(best_x)
        _move(V2, V1, x)
        remaining.remove(x)
        
        if ((minimize and best_obj_local < best_obj) or
            (not minimize and best_obj_local > best_obj)):
            best_obj = best_obj_local
            best_bp = [(list(V1), list(V2))]
        elif best_obj_local == best_obj:
            best_bp.append( (list(V1), list(V2)) )
            
    return best_obj, random.choice(best_bp)


# ----------------------------------------------------------------------------
#                        Adaptive walk bipartition
# ----------------------------------------------------------------------------

def gradient_walk_bipartition(V, f_obj, 
                              minimize=True, args=(),
                              initial_bipartition=None):
    
    if len(V) < 2:
        raise ValueError('iterable must have >=2 elements')
    
    # partition P
    P = [set(), set()]
    lookup = {}
    
    if not isinstance(V, (list, tuple)):
        V = list(V)
    
    if not initial_bipartition:
        for k, m in enumerate(np.random.permutation(len(V))):
            x = V[m]
            i = 0 if k <= len(V) / 2 else 1
            P[i].add(x)
            lookup[x] = i
    elif len(initial_bipartition) == 2:
        for i in range(2):
            for x in initial_bipartition[i]:
                P[i].add(x)
                lookup[x] = i
    else:
        raise ValueError('not a bipartition')
        
    best_obj = f_obj([*P], *args)
    best_bp = [list(s) for s in P]
    
    while True:
        
        best_obj_local = float('inf') if minimize else float('-inf')
        best_x = None
        
        for x in V:
            V1 = P[lookup[x]]
            V2 = P[(lookup[x] + 1) % 2]
            
            # avoid that one set becomes empty
            if len(V1) == 1:
                continue
            
            _move(V1, V2, x)
            obj = f_obj([V1, V2], *args)
            if ((minimize and obj < best_obj_local) or
                (not minimize and obj > best_obj_local)):
                best_obj_local = obj
                best_x = [x]
            elif obj == best_obj_local:
                best_x.append(x) 
            _move(V2, V1, x)
        
        if ((minimize and best_obj_local < best_obj) or
            (not minimize and best_obj_local > best_obj)):
            x = random.choice(best_x)
            V1 = P[lookup[x]]
            V2 = P[(lookup[x] + 1) % 2]
            _move(V1, V2, x)
            lookup[x] = (lookup[x] + 1) % 2
            
            best_obj = best_obj_local
            best_bp = [list(s) for s in P]
        else:
            break
    
    return best_obj, best_bp
        
    

if __name__ == '__main__':
    
    import networkx as nx
    
    G = nx.Graph()
    G.add_edges_from([('a', 'b'), ('a', 'c'), ('a', 'd'),
                      ('b', 'd'), ('c', 'd'), ('c', 'e'), ('d', 'e')])
        
    V = list(G.nodes())
    
    for i in range(3):
        print( *greedy_bipartition(V, partition_cut_value, args=(G,)) )
    
    print('---------')
    for i in range(3):
        print( *gradient_walk_bipartition(V, partition_cut_value, args=(G,)) )




# -*- coding: utf-8 -*-

import networkx as nx

from tralda.datastructures.Tree import Tree, TreeNode
from tralda.supertree.Build import aho_graph, mtt_partition

from bmgedit.partitioning.Karger import Karger
from bmgedit.partitioning.GreedyBipartition import (greedy_bipartition,
                                                    gradient_walk_bipartition)
from bmgedit.partitioning.Louvain import Louvain, LouvainCustomObj
from bmgedit.partitioning.NumberPartition import balanced_coarse_graining


__author__ = 'David Schaller'


class Build2:
    """BUILD / MTT algorithm with optimal objective partition."""
    
    def __init__(self, R, L, F=None,
                 allow_inconsistency=True,
                 binarize=False,
                 part_method='mincut',
                 obj_function=None,
                 minimize=True,
                 obj_function_args=None,
                 greedy_repeats=5,
                 weighted_mincut=False, triple_weights=None,):
        
        self.R = R
        self.L = L
        
        # forbidden triples --> activates MTT if non-empty
        self.F = F
        
        # allow inconsistencies or return False?
        self.allow_inconsistency = allow_inconsistency
        
        # mode for tree binarization
        if not binarize:
            self.binarize = False
        elif isinstance(binarize, str) and \
            binarize.lower() in ('balanced', 'b'):
            self.binarize = 'b'
        elif isinstance(binarize, str) and \
            binarize.lower() in ('caterpillar', 'c'):
            self.binarize = 'c'
        else:
            raise ValueError('unknown binarization mode:' \
                             ' {}'.format(binarize))
        
        # partitioning method in case of inconsistencies
        if part_method in ('mincut', 'karger', 'greedy', 'gradient_walk',
                           'louvain', 'louvain_obj'):
            self.part_method = part_method
        else:
            raise ValueError("unknown bipartition method "\
                             "'{}'".format(part_method))
        self.obj_function =      obj_function
        self.minimize =          minimize
        self.obj_function_args = obj_function_args
        self.greedy_repeats =    greedy_repeats
            
        # parameters if bipartition method is mincut
        self.weighted_mincut = weighted_mincut
        self.triple_weights = triple_weights
    
    
    def build_tree(self, return_root=False):
        """Build a tree displaying all triples in R if possible.
        
        Keyword arguments:
            return_root - if True, return 'TreeNode' instead of
                'Tree' instance
        """
        
        self.total_obj = 0
        
        if self.F:
            root = self._mtt(self.L, self.R, self.F)
        else:
            root = self._aho(self.L, self.R)
            
        return root if return_root else Tree(root)
    
    
    def _trivial_case(self, L):
        
        if len(L) == 1:
            leaf = L.pop()
            return TreeNode(label=leaf)
        
        elif len(L) == 2:
            node = TreeNode()
            for _ in range(2):
                leaf = L.pop()
                node.add_child(TreeNode(label=leaf))
            return node
    
    
    def _aho(self, L, R):
        """Recursive Aho-algorithm."""
        
        # trivial case: one or two leaves left in L
        if len(L) <= 2:
            return self._trivial_case(L)
            
        aux_graph = aho_graph(R, L, weighted=self.weighted_mincut,
                                    triple_weights=self.triple_weights)
        part = list(nx.connected_components(aux_graph))
        
        if len(part) < 2:
            if not self.allow_inconsistency:
                return False
            else:
                obj, part = partition(L, self.part_method,
                                      obj_function=self.obj_function,
                                      minimize=self.minimize,
                                      args=self.obj_function_args,
                                      aux_graph=aux_graph,
                                      greedy_repeats=self.greedy_repeats)
                self.total_obj += obj
        
        if self.binarize == 'b' and len(part) > 2:
            part = balanced_coarse_graining(part)
        
        root = TreeNode()                   # place new inner node
        node = root
        for i, s in enumerate(part):
            Li, Ri = set(s), []
            for t in R:                     # construct triple subset
                if Li.issuperset(t):
                    Ri.append(t)
            Ti = self._aho(Li, Ri)          # recursive call
            if not Ti:
                return False                # raise False to previous call
            else:
                node.add_child(Ti)          # add roots of the subtrees
            
            # resolve to binary (caterpillar)
            if self.binarize == 'c' and i < len(part)-2:
                new_node = TreeNode()
                node.add_child(new_node)
                node = new_node
   
        return root
    
    
    def _mtt(self, L, R, F):
        """Recursive MTT algorithm."""
        
        # trivial case: one or two leaves left in L
        if len(L) <= 2:
            return self._trivial_case(L)
        
        part, aux_graph = mtt_partition(L, R, F)
        
        if len(part) < 2:
            if not self.allow_inconsistency:
                return False
            else:
                obj, part = partition(L, self.part_method,
                                      obj_function=self.obj_function,
                                      minimize=self.minimize,
                                      args=self.obj_function_args,
                                      aux_graph=aux_graph,
                                      greedy_repeats=self.greedy_repeats)
                self.total_obj += obj
        
        if self.binarize == 'b' and len(part) > 2:
            part = balanced_coarse_graining(part)
        
        root = TreeNode()                   # place new inner node
        node = root
        for i, s in enumerate(part):
            Li, Ri, Fi = set(s), [], []
            for Xi, X in ((Ri, R), (Fi, F)):
                for t in X:
                    if Li.issuperset(t):
                        Xi.append(t)
            Ti = self._mtt(Li, Ri, Fi)      # recursive call
            if not Ti:
                return False                # raise False to previous call
            else:
                node.add_child(Ti)          # add roots of the subtrees
            
            # resolve to binary (caterpillar)
            if self.binarize == 'c' and i < len(part)-2:
                new_node = TreeNode()
                node.add_child(new_node)
                node = new_node
   
        return root
    
    
def partition(L, method,
              obj_function=None, minimize=True, args=None,
              aux_graph=None,
              greedy_repeats=1):
    
    best_obj = float('inf') if minimize else float('-inf')
    best_part = None
    
    if method == 'mincut':
        
        # Stoer–Wagner algorithm
        best_obj, best_part = nx.stoer_wagner(aux_graph)
    
    elif method == 'karger':
        karger = Karger(aux_graph)
        
        for _, part in karger.generate(runs=greedy_repeats):
            obj = obj_function(part, *args)
            
            if ((minimize and obj < best_obj) or
                (not minimize and obj > best_obj)):
                best_obj, best_part = obj, part
    
    elif method == 'greedy':
        
        for _ in range(greedy_repeats):
            obj, part = greedy_bipartition(L, obj_function, 
                                           minimize=minimize,
                                           args=args)
            if ((minimize and obj < best_obj) or
                (not minimize and obj > best_obj)):
                best_obj, best_part = obj, part
    
    elif method == 'gradient_walk':
        
        for _ in range(greedy_repeats):
            obj, part = gradient_walk_bipartition(L, obj_function,
                                                  minimize=minimize,
                                                  args=args)
            if ((minimize and obj < best_obj) or
                (not minimize and obj > best_obj)):
                best_obj, best_part = obj, part
    
    elif method == 'louvain':
        
        best_obj = float('-inf')    # modularity is always maximized
        
        for _ in range(greedy_repeats):
            louv = Louvain(aux_graph, at_least_two=True)
            obj, part = louv.modularities[-1], louv.partitions[-1]
            if obj > best_obj:
                best_obj, best_part = obj, part
    
    elif method == 'louvain_obj':
        
        for _ in range(greedy_repeats):
            louv = LouvainCustomObj(aux_graph, obj_function,
                                    minimize=minimize,
                                    args=args,
                                    at_least_two=True)
            obj, part = louv.objectives[-1], louv.partitions[-1]
            if ((minimize and obj < best_obj) or
                (not minimize and obj > best_obj)):
                best_obj, best_part = obj, part
    
    return best_obj, best_part



# -*- coding: utf-8 -*-

"""
Scenario.

Wrapper class for species and gene tree. Compute statistics, BMG/RBMG.
"""

import networkx as nx

from asymmetree.treeevolve.GeneTree import observable_tree
from asymmetree.analysis.BestMatches import orthology_from_tree, bmg_from_tree
from asymmetree.tools.PhyloTreeTools import (reconc_sorted_leaves,
                                             distance_matrix,)


__author__ = 'David Schaller'


class Scenario:
    
    def __init__(self, S, TGT, dupl_rate, loss_rate, hgt_rate, OGT=None):
        
        self.S = S
        self.TGT = TGT
        self.DLH_rates = (dupl_rate, loss_rate, hgt_rate)
        
        self.OGT = OGT if OGT else observable_tree(TGT)
        
        self.reconc_dict, self.genes = reconc_sorted_leaves(self.OGT, return_list=True)
        self.gene_index = {gene: i for i, gene in enumerate(self.genes)}
        
        self._count_events()
        self._sort_species_to_subtrees()
        
        self.TOG = orthology_from_tree(self.OGT)
        self.bmg, self.rbmg = bmg_from_tree(self.OGT, supply_rbmg=True)
        
        self.bmg_subtrees, self.rbmg_subtrees = self.reduce_to_subtrees(self.bmg, self.rbmg)
    
    
    def _count_events(self):
        
        self.event_counts = [self.S.number_of_species, 0, 0, 0, 0]    # count S / D / L / H /
                                                                      # ancient duplications                
        for v in self.TGT.preorder():
            if v.event == 'D':
                self.event_counts[1] += 1
                if self.S.root.label == v.reconc[0]:
                    self.event_counts[4] += 1
            elif v.event == 'L':
                self.event_counts[2] += 1
            elif v.event == 'H':
                self.event_counts[3] += 1
    
    
    def _sort_species_to_subtrees(self):
        
        # sort the species into the subtrees of the first speciation event of S
        self.subtree_list, self.subtree_index = [], {}
        self.S_leaves = self.S.leaf_dict()
        for u in self.S.preorder():                          # exclude root of planted tree
            if len(u.children) > 1:
                for i in range(len(u.children)):
                    self.subtree_list.append([l.label for l in self.S_leaves[u.children[i]]])
                    self.subtree_index.update({item: i for item in self.subtree_list[i]})
                break
        
        # max. number of outgroup genes (in T) for each subtree of S
        self.outgroup_dict = {i: [] for i in range(len(self.subtree_list))}
        for gene in self.genes:
            for i in self.outgroup_dict.keys():
                if self.subtree_index[gene.reconc] != i:
                    self.outgroup_dict[i].append(gene)
    
    
    def get_data(self):
        
        return (self.S, self.genes, self.gene_index,
                self.subtree_list, self.subtree_index,
                self.outgroup_dict, self.reconc_dict)
    
    
    def rates_and_counts(self):
        """Return event rates and counts."""
        
        return list(self.DLH_rates) + self.event_counts
    
    
    def reduce_to_subtrees(self, full_bmg, full_rbmg):
        """Return a subgraph of the true RBMG with edges {u,v} for which the
        corresponding species are in the same subtree of root(S)."""
        
        bmg_subtrees = nx.DiGraph()
        rbmg_subtrees = nx.Graph()
        
        for G_subtrees, true_G in [(bmg_subtrees, full_bmg),
                                   (rbmg_subtrees, full_rbmg)]:
            G_subtrees.add_nodes_from(true_G.nodes(data=True))
            for u, v in true_G.edges:
                u_col = true_G.nodes[u]['color']
                v_col = true_G.nodes[v]['color']
                
                if self.subtree_index[u_col] == self.subtree_index[v_col]:
                    G_subtrees.add_edge(u, v)
        
        return bmg_subtrees, rbmg_subtrees
    
    
    def get_distance_matrix(self):
        
        _, D = distance_matrix(self.OGT, leaf_order=self.genes)
        
        return D
    
    
    def possible_edges_bmg(self):
        """Return the number of possible edges in the BMG i.e. for gene pairs for
        which an outgroup is available in S."""
        
        counts = [0 for i in range(len(self.subtree_list))]
        
        for g in self.genes:
            counts[self.subtree_index[g.reconc]] += 1
        
        possible_edges = 0
        for count in counts:
            possible_edges += (count * (count-1))
        
        return possible_edges



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='bmgedit',
    version='0.0.5',
    author='David Schaller',
    author_email='sdavid@bioinf.uni-leipzig.de',
    description='Best match graph editing.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/david-schaller/bmg-edit',
    packages=setuptools.find_packages('src'),
    package_dir={'': 'src'},
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.7',
    install_requires=[
        'numpy>=1.16.4',
        'scipy>=1.3.0',
        'matplotlib>=3.0',
        'networkx>=2.2',
        'asymmetree>=2.2.0',
        'tralda>=1.0.1',
   ],
)



# -*- coding: utf-8 -*-

"""Heuristics for BMG editing."""

import itertools

from tralda.datastructures import LCA
from tralda.supertree.Build import best_pair_merge_first
from tralda.tools.GraphTools import sort_by_colors

from asymmetree.analysis.BestMatches import (informative_triples,
                                             binary_explainable_triples,
                                             bmg_from_tree)
from asymmetree.tools.PhyloTreeTools import (reconstruct_reconc_from_graph,)


from bmgedit.Build import Build2


__author__ = 'David Schaller'


class BMGEditor:
    """Wrapper for triple-based BMG editing heuristics."""
    
    def __init__(self, G, binary=False, binarization_mode='balanced', use_binary_triples=True):
        
        self.G = G
        self.color_dict = sort_by_colors(G)
        
        self.L = [v for v in G.nodes()]
        
        # informative triples
        if not binary:
            self.binarize = False
            if use_binary_triples:
                self.R = binary_explainable_triples(G, color_dict=self.color_dict)
            else:
                self.R = informative_triples(G, color_dict=self.color_dict)
        else:
            self.binarize = binarization_mode
            self.R = binary_explainable_triples(G, color_dict=self.color_dict)
            
        # current tree built by one of the heuristics
        self._tree = None


    def extract_consistent_triples(self):
        
        if not self._tree:
            raise RuntimeError('no tree has been built yet')
        
        lca = LCA(self._tree)
        
        return lca.consistent_triples(self.R)
    
    
    def build(self, method, objective='cost'):
        
        if objective == 'cost':
            minimize = True
            f_obj = unsatisfiability_cost
        elif objective == 'gain':
            minimize = False
            f_obj = satisfied_relations
        else:
            raise ValueError('unknown mode for objective ' \
                             'function: {}'.format(objective))

        method = method.lower()
        
        if method == 'bpmf':
            self._tree = best_pair_merge_first(self.R,self.L,
                                               triple_weights=None) 
        elif method in ('mincut', 'karger', 'greedy', 'gradient_walk',
                        'louvain', 'louvain_obj'):
            build = Build2(self.R, self.L,
                           allow_inconsistency=True,
                           binarize=self.binarize,
                           part_method=method,
                           obj_function=f_obj,
                           minimize=minimize,
                           obj_function_args=(self.G,),
                           weighted_mincut=True,
                          )
            self._tree = build.build_tree()
        else:
            raise ValueError('unknown partition method: {}'.format(method))
    
    
    def get_bmg(self, extract_triples_first=False,
                supply_inner_vertex_count=False,
                update_tree= False,
               ):
        
        if not extract_triples_first:
            reconstruct_reconc_from_graph(self._tree, self.G)
            tree = self._tree
        else:
            R_consistent = self.extract_consistent_triples()
            build = Build2(R_consistent, self.L,
                           allow_inconsistency=False,
                           binarize=self.binarize,
                          )
            tree = build.build_tree()
            reconstruct_reconc_from_graph(tree, self.G)

            if update_tree:
                self._tree= tree

        if supply_inner_vertex_count:
            return bmg_from_tree(tree), sum(1 for _ in tree.inner_nodes())
        else:
            return bmg_from_tree(tree)
        
        
def unsatisfiability_cost(partition, G):
    """Return the unsatisfiability cost of a partition.
    
    This cost is the number of unsatisfiable (non-)arcs induced by the
    bipartition.
    
    Parameters
    ----------
    partition : iterable of iterables of nodes
        A partition for (a subset of) the nodes in the graph.
    G : networkx.DiGraph
        A directed graph.
    
    Returns
    -------
    int
        The number of unsatisfiable best match arcs and non-arcs between
        elements in the partition, if this partition correspond to a split
        in a tree.
    """
    
    partition = list(partition)
    color_sets = [{} for V in partition]
    
    for V, colors in zip(partition, color_sets):
        for v in V:
            c = G.nodes[v]['color']
            if c not in colors:
                colors[c] = 1
            else:
                colors[c] += 1
    
    cost = 0
    
    for V, colors in zip(partition, color_sets):
        
        if not isinstance(V, set):
            V = set(V)
        
        for x, y in itertools.product(V, G.nodes()):
            
            y_color = G.nodes[y]['color']
            
            # skip pairs with the same color
            if G.nodes[x]['color'] == y_color:
                continue
            
            if y not in V:
                if G.has_edge(x, y):
                    if colors.get(y_color):
                        cost += 1
                elif not colors.get(y_color):
                    cost += 1
            elif not G.has_edge(x, y) and colors.get(y_color) == 1:
                cost += 1
                    
    return cost


def get_U1_U2_U3(partition, G):
    """Get the unsatisfiable relations of each type for a partition.
    
    U1: (x, y) in E,
        x in V_i, y in V \ V_i,
        color of y is present in V_i
    U2: (x, y) not in E,
        x in V_i, y in V \ V_i,
        color of y is not present in V_i
    U3: (x, y) not in E,
        distinct x, y in V_i
        y is the only vertex of its color in V_i.
    
    Parameters
    ----------
    partition : iterable of iterables of nodes
        A partition for (a subset of) the nodes in the graph.
    G : networkx.DiGraph
        A directed graph.
    
    Returns
    -------
    tuple of three lists of arcs
        The unsatisfiable arcs and non-arcs if the partition corresponds to a
        split in a tree sorted by the three types.
    """
    
    partition = list(partition)
    color_sets = [{} for V in partition]
    
    for V, colors in zip(partition, color_sets):
        for v in V:
            c = G.nodes[v]['color']
            if c not in colors:
                colors[c] = 1
            else:
                colors[c] += 1
    
    U1, U2, U3 = [], [], []
    
    for V, colors in zip(partition, color_sets):
        
        if not isinstance(V, set):
            V = set(V)
        
        for x, y in itertools.product(V, G.nodes()):
            
            y_color = G.nodes[y]['color']
            
            # skip pairs with the same color
            if G.nodes[x]['color'] == y_color:
                continue
            
            if y not in V:
                if G.has_edge(x, y):
                    if colors.get(y_color):
                        U1.append( (x, y) )
                elif not colors.get(y_color):
                    U2.append( (x, y) )
            elif not G.has_edge(x, y) and colors.get(y_color) == 1:
                U3.append( (x, y) )
                    
    return U1, U2, U3


def satisfied_relations(partition, G):
    """Return the no. of arcs and non-arcs that are satisfied by a partition.
    
    Only considers pairs of vertices of different color.
    
    Parameters
    ----------
    partition : iterable of iterables of nodes
        A partition for (a subset of) the nodes in the graph.
    G : networkx.DiGraph
        A directed graph.
    
    Returns
    -------
    int
        The number of best match arcs and non-arcs between vertices of
        different colors that are satisfied if the partition corresponds to
        a split in a leaf-colored tree.
    """
    
    partition = list(partition)
    color_sets = [{} for V in partition]
    
    for V, colors in zip(partition, color_sets):
        for v in V:
            c = G.nodes[v]['color']
            if c not in colors:
                colors[c] = 1
            else:
                colors[c] += 1
    
    count = 0
    
    for V, colors in zip(partition, color_sets):
        
        if not isinstance(V, set):
            V = set(V)
        
        for x, y in itertools.product(V, G.nodes()):
            
            y_color = G.nodes[y]['color']
            
            # skip pairs with the same color
            if G.nodes[x]['color'] == y_color:
                continue
            
            if y not in V:
                if not G.has_edge(x, y):
                    if colors.get(y_color):
                        count += 1
                elif not colors.get(y_color):
                    count += 1
            elif G.has_edge(x, y) and colors.get(y_color) == 1:
                count += 1
                    
    return count


def get_S1_S2_S3(partition, G):
    """Get the arcs and non-arcs that are satisfied by a partition of each type.
    
    S1: (x, y) notin E,
        x in V_i, y in V \ V_i,
        color of y is present in V_i
    S2: (x, y) in E,
        x in V_i, y in V \ V_i,
        color of y is not present in V_i
    S3: (x, y) in E,
        distinct x, y in V_i
        y is the only vertex of its color in V_i.
    
    Parameters
    ----------
    partition : iterable of iterables of nodes
        A partition for (a subset of) the nodes in the graph.
    G : networkx.DiGraph
        A directed graph.
    
    Returns
    -------
    tuple of three lists of arcs
        The arcs and non-arcs that are satisfied by a partition if this
        partition corresponds to a split in a tree sorted by the three types.
    """
    
    partition = list(partition)
    color_sets = [{} for V in partition]
    
    for V, colors in zip(partition, color_sets):
        for v in V:
            c = G.nodes[v]['color']
            if c not in colors:
                colors[c] = 1
            else:
                colors[c] += 1
    
    S1, S2, S3 = [], [], []
    
    for V, colors in zip(partition, color_sets):
        
        if not isinstance(V, set):
            V = set(V)
        
        for x, y in itertools.product(V, G.nodes()):
            
            y_color = G.nodes[y]['color']
            
            # skip pairs with the same color
            if G.nodes[x]['color'] == y_color:
                continue
            
            if y not in V:
                if not G.has_edge(x, y):
                    if colors.get(y_color):
                        S1.append( (x, y) )
                elif not colors.get(y_color):
                    S2.append( (x, y) )
            elif G.has_edge(x, y) and colors.get(y_color) == 1:
                S3.append( (x, y) )
                    
    return S1, S2, S3


if __name__ == '__main__':
    
    from time import time
    
    from tralda.tools.GraphTools import disturb_graph, symmetric_diff
    
    from asymmetree.tools.PhyloTreeTools import random_colored_tree
    from asymmetree.analysis.BestMatches import is_bmg
    
    random_tree = random_colored_tree(30, 10, force_all_colors=True)
    bmg = bmg_from_tree(random_tree)
    disturbed = disturb_graph(bmg, 0.1, 0.1, preserve_properly_colored=True)
    
    print('-------------')
    print([(v, bmg.nodes[v]['color']) for v in bmg.nodes()])
    print('orginal', bmg.edges(), bmg.size())
    print('disturbed', disturbed.edges(), disturbed.size())
    print('-------------')
    print('original BMG is BMG:   ', bool(is_bmg(bmg)  ) )
    print('disturbed graph is BMG:', bool(is_bmg(disturbed)) )
    print('original vs disturbed:', symmetric_diff(bmg, disturbed))
    print('-------------')
    
    editor = BMGEditor(disturbed, binary=True)
    for method in ('Mincut', 'BPMF', 'Karger',
                   'Greedy', 'Gradient_Walk',
                   'Louvain', 'Louvain_Obj'):
        for objective in ('cost', 'gain'):
            start_time = time()
            editor.build(method, objective=objective)
            end_time = time() - start_time
            bmg1 = editor.get_bmg(extract_triples_first=False)
            bmg2 = editor.get_bmg(extract_triples_first=True)
            print('-----', method, objective, '-----')
            print('time:', end_time)
            print('orig./dist. vs BMG1 (no extr):',
                  symmetric_diff(bmg, bmg1), symmetric_diff(disturbed, bmg1))
            print('orig./dist. vs BMG2 (extr):',
                  symmetric_diff(bmg, bmg2), symmetric_diff(disturbed, bmg2))



# -*- coding: utf-8 -*-

"""
Quartet approach for best match inference.
"""

import itertools, random,  subprocess, time, os

import numpy as np
import networkx as nx

from tralda.datastructures.Tree import LCA

from bmg_edit.best_match_infer.ScenarioFileIO import (parse_bmg_edges,
                                                      matrix_to_phylip,
                                                      species_to_genes,
                                                      write_newick)


__author__ = 'David Schaller'


# --------------------------------------------------------------------------
#                           PYTHON IMPLEMENTATION
#
# --------------------------------------------------------------------------
# Outgroup selection methods:
#  (1) outgroups from subtrees under the root of S
#  (2) use all possible outgoups for a pair of species
#      (corrected closest outgroups)
# --------------------------------------------------------------------------

class Quartets:
    """Implementation of the Quartet approach for best match inference."""
    
    def __init__(self, scenario, distance_matrix,
                 voting_mode='majority',
                 use_distant_genes=False,
                 insecurity_factor=0.0,
                 y_candidates=None,
                 closest_outgroups=False,
                 incongruence_threshold=0.2):
        """Inititialize Quartet method.
        
        Keyword arguments:
            voting_mode -- 'majority' or 'weighted' voting for quartet
                           support.
            use_distant_genes -- use the most distant outgroups instead
                                 of a random choice (only root outgroups)
            insecurity_factor -- threshold for inference of the star
                                 topology
            y_candidates -- restrict the Y candidates (dict x: Y)
            closest_outgroups -- apply the corrected closest outgroups
                                 method
            incongruence_threshold -- threshold for discarding
                                      outgroup species
        """
        
        self.scenario = scenario
        
        (self.S, self.genes, self.gene_index,
         self.subtree_list, self.subtree_index,
         self.outgroup_dict, self.reconc_dict) = scenario.get_data()
        
        self.D = distance_matrix
        self.voting_mode = voting_mode                          # majority voting or weighted sum
        self.use_distant_genes = use_distant_genes              # use most distant outgroups instead of random choice
        self.insecurity_factor = insecurity_factor              # relative distance between top votes
                                                                # to decide for star topology instead
        
        self.y_candidates = y_candidates                        # Graph or DiGraph that contains
                                                                # best match candidates
        
        self.closest_outgroups = closest_outgroups              # use the closest corrected outgroups method
        self.incongruence_threshold = incongruence_threshold    # threshold ancient dupl. correction
    
    
    def compute_outgroup_species(self):
        
        self.lca_to_outgroups = {key: set() for key in self.subtree_species}
        
        root_done = False
        for v in self.S.preorder():
            
            if v.parent:
                self.lca_to_outgroups[v.label].update(self.lca_to_outgroups[v.parent.label])
                
            for c1, c2 in itertools.permutations(v.children, 2):
                
                if self.incongruence_threshold >= 0 and root_done:
                    outgroup_candidates = set(self.subtree_species[c2.label])
                    genes_c1 = self.subtree_genes[c1.label]
                    species_c2 = self.subtree_species[c2.label]
                    if len(genes_c1) > 1 and len(species_c2) > 1:
                        for s1, s2 in itertools.combinations(species_c2, 2):
                            genes_c2 = self.reconc_dict[s1] + self.reconc_dict[s2]
                            
                            votes = [0,0]                                           # [congruent, incongruent]
                            for i in range(20):
                                a, b = random.sample(genes_c1, 2)
                                c, d = random.sample(genes_c2, 2)
                                quartet, _ = self.supported_quartet(a, b, c, d)
                                if quartet == 0:                                    # ab | cd
                                    votes[0] += 1                                   # (congruent)
                                else:                                               # ac | bd or ad | bc or star
                                    votes[1] += 1                                   # (incongruent)
                                    
                            if votes[1] / sum(votes) >= self.incongruence_threshold:
                                outgroup_candidates.discard(s1)
                                outgroup_candidates.discard(s2)
                            
                    self.lca_to_outgroups[c1.label].update(outgroup_candidates)
                    
                else:                                               # outgroups on the root are fix
                    self.lca_to_outgroups[c1.label].update(self.subtree_species[c2.label])
                root_done = True

    
    def build_matrix_I(self):
        """Build an index matrix that stores for each gene (row) the indeces
        of the other genes in increasing order by distance."""
        
        N = self.D.shape[0]
        self.I = np.tile(np.arange(N), (N,1))
        D, I = self.D, self.I
        
        def quicksort_row(row, l, r):
            if r <= l:
                return
            i, j = l, r
            piv = D[row, I[row, (l+r)//2]]
            while i <= j:
                while D[row, I[row, i]] < piv:
                    i += 1
                while D[row, I[row, j]] > piv:
                    j -= 1
                if i <= j:
                    I[row,i], I[row,j] = I[row,j], I[row,i]
                    i += 1
                    j -= 1
            quicksort_row(row, l, j)
            quicksort_row(row, i, r)
        
        for row in range(N):
            quicksort_row(row, 0, N-1)
        
    
    def get_outgroups_root_only(self, x, limit=10):
        
        i, x_subtree = self.gene_index[x], self.subtree_index[x.reconc]
        limit = min(limit, len(self.outgroup_dict[x_subtree]))
        
        if self.use_distant_genes:
            outgroups = []
            j = len(self.genes)-1
            while len(outgroups) < limit and j >= 0:
                gene = self.genes[self.I[i,j]]
                if x_subtree != self.subtree_index[gene.reconc]:
                    outgroups.append(gene)
                j -= 1
        else:
            outgroups = random.sample(self.outgroup_dict[x_subtree], limit)
        
        return outgroups
    
    
    def get_outgroups_closest(self, x, reconcY, limit=10):
        
        outgroups_S = self.lca_to_outgroups[self.lca_S(x.reconc, reconcY).label]
        
        outgroups = []
        i, j, N = self.gene_index[x], 0, self.D.shape[0]
        while len(outgroups) < limit and j < N:
            if self.genes[self.I[i,j]].reconc in outgroups_S:
                outgroups.append(self.genes[self.I[i,j]])
            j += 1
        
        return outgroups
    
    
    def supported_quartet(self, x, y1, y2, z):
        
        x, y1, y2, z = (self.gene_index[x], self.gene_index[y1],
                        self.gene_index[y2], self.gene_index[z])
        
        xy1_y2z = self.D[x, y1] + self.D[y2, z ]
        xy2_y1z = self.D[x, y2] + self.D[y1, z ]
        xz_y1y2 = self.D[x, z ] + self.D[y1, y2]
        
        if xy1_y2z < xy2_y1z and xy1_y2z < xz_y1y2:     # 0: xy1 | y2z
            quartet = 0
        elif xy2_y1z < xy1_y2z and xy2_y1z < xz_y1y2:   # 1: xy2 | y1z
            quartet = 1
        elif xz_y1y2 < xy1_y2z and xz_y1y2 < xy2_y1z:   # 2: xz | y1y2
            quartet = 2
        else:                                           # 3: star
            quartet = 3
            
        d0, d1, d2 = sorted([xy1_y2z, xy2_y1z, xz_y1y2])
        weight = (1 - d0/d2) * np.exp(d1 - d2) if d2 > 0 else 0.0
        
        return quartet, weight
    
    
    def find_best_matches(self, x, Y, Z):
        
        H = nx.DiGraph()
        f = 1 - self.insecurity_factor
        
        for y1, y2 in itertools.combinations(Y, 2):
            votes = [0, 0, 0, 0]
            for z in Z:
                q, weight = self.supported_quartet(x, y1, y2, z)
                if self.voting_mode == "weighted sum":
                    votes[q] += weight                          # use the weight
                else:
                    votes[q] += 1                               # majority vote
            
            sorted_votes = sorted((vote, i) for i, vote in enumerate(votes))
            
            if sorted_votes[2][0] > f * sorted_votes[3][0]:     # not sure if the two 
                quartet = 3                                     # top votes are too close
                                                                # default is star topology
            else:
                quartet = sorted_votes[3][1]
            
            if quartet == 0:                                    # 0: xy1 | y2z
                H.add_edge(y2, y1)
            elif quartet == 1:                                  # 1: xy2 | y1z
                H.add_edge(y1, y2)
            else:                                               # 2: xz | y1y2   or   3: star
                H.add_edge(y1, y2)
                H.add_edge(y2, y1)
        
        # find the (unique) strongly connected component without out-edges
        sccs = {i: scc for i, scc in enumerate(nx.strongly_connected_components(H))}
        
        for i, scc in sccs.items():
            for y in scc:
                H.nodes[y]["scc"] = i
        while sccs:
            for i, scc in sccs.items():
                out_edges = False
                for y1 in scc:
                    for y2 in H.successors(y1):
                        if H.nodes[y1]["scc"] != H.nodes[y2]["scc"]:
                            out_edges = True
                            break
                    if out_edges:
                        break
                if not out_edges:
                    return list(scc)
    
    
    def build_graphs_root_only(self):
        
        if self.use_distant_genes:
            self.build_matrix_I()  
        
        self.bmg = nx.DiGraph()
        for g in self.genes:
            self.bmg.add_node(g.label, color=g.reconc)
        
        # consider all genes of other species as potential best matches
        if self.y_candidates is None:
            
            for x in self.genes:
                Z = self.get_outgroups_root_only(x)
                
                for species in self.subtree_list[self.subtree_index[x.reconc]]:
                    Y = self.reconc_dict[species]
                    if species == x.reconc:
                        continue
                    elif len(Y) == 1:
                        self.bmg.add_edge(x.label, Y[0].label)
                    elif len(Y) > 1:
                        best_matches = self.find_best_matches(x, Y, Z)
                        for y in best_matches:
                            self.bmg.add_edge(x.label, y.label)
                            
        # consider only a given subset as best matches
        else:
            id_dict = {g.label: g for g in self.genes}
            for x in self.genes:
                Z = self.get_outgroups_root_only(x)
                Y = sorted([id_dict[gene_id] for gene_id in self.y_candidates.neighbors(x.label)],
                           key=lambda y: y.reconc)               # sort by reconc
                
                i = 0
                while i < len(Y):                               # handle reconcs separately
                    j = i
                    while j < len(Y) and Y[i].reconc == Y[j].reconc:
                        j += 1
                    if self.subtree_index[x.reconc] == self.subtree_index[Y[i].reconc]:
                        if i == j-1:
                            self.bmg.add_edge(x.label, Y[i].label)
                        else:
                            best_matches = self.find_best_matches(x, Y[i:j], Z)
                            for y in best_matches:
                                self.bmg.add_edge(x.label, y.label)
                    i = j
        
        return self.bmg, self.bmg.to_undirected(reciprocal=True)
    
    
    def build_graphs_closest_outgroups(self):
        
        self.lca_S = LCA(self.S)
        self.S_leaves = self.S.leaf_dict()
        
        self.subtree_species = {}
        self.subtree_genes = {}
        for v in self.S.preorder():
            self.subtree_species[v.label] = [leaf.label for leaf in self.S_leaves[v]]
            self.subtree_genes[v.label] = []
            for leaf_id in self.subtree_species[v.label]:
                self.subtree_genes[v.label].extend(self.reconc_dict[leaf_id])
        
        self.compute_outgroup_species()
        self.build_matrix_I()
        
        self.bmg = nx.DiGraph()
        for g in self.genes:
            self.bmg.add_node(g.label, color=g.reconc)
        
        # consider all genes of other species as potential best matches
        if self.y_candidates is None:
            
            for x in self.genes:
                for species in self.subtree_list[self.subtree_index[x.reconc]]:
                    Y = self.reconc_dict[species]
                    if species == x.reconc:
                        continue
                    elif len(Y) == 1:
                        self.bmg.add_edge(x.label, Y[0].label)
                    elif len(Y) > 1:
                        Z = self.get_outgroups_closest(x, Y[0].reconc)
                        best_matches = self.find_best_matches(x, Y, Z)
                        for y in best_matches:
                            self.bmg.add_edge(x.label, y.label)
                            
        # consider only a given subset as best matches
        else:
            id_dict = {g.label: g for g in self.genes}
            for x in self.genes:
                Y = sorted([id_dict[gene_id] for gene_id in self.y_candidates.neighbors(x.label)],
                           key=lambda y: y.reconc)               # sort by reconc
                
                i = 0
                while i < len(Y):                               # handle species separately
                    j = i
                    while j < len(Y) and Y[i].reconc == Y[j].reconc:
                        j += 1
                    if self.subtree_index[x.reconc] == self.subtree_index[Y[i].reconc]:
                        if i == j-1:
                            self.bmg.add_edge(x.label, Y[i].label)
                        else:
                            Z = self.get_outgroups_closest(x, Y[i].reconc)
                            best_matches = self.find_best_matches(x, Y[i:j], Z)
                            for y in best_matches:
                                self.bmg.add_edge(x.label, y.label)
                    i = j
                    
        return self.bmg, self.bmg.to_undirected(reciprocal=True)
    
    
    def build_graphs(self):
        
        if not self.closest_outgroups:
            self.build_graphs_root_only()
        else:
            self.build_graphs_closest_outgroups()
            


# --------------------------------------------------------------------------
#                            EXTERNAL C++ PROGRAM
#
# --------------------------------------------------------------------------

def quartet_qinfer(scenario,
                   matrix_filename, species_filename, tree_filename,
                   epsilon=-1,
                   closest_outgroups=False,
                   incongruence_threshold=0.2,
                   benchmark_file=None,
                   binary_path=None):
    """Compute BMG and RBMG from a distances matrix D using 'qinfer'.
    
    Keyword arguments:
        epsilon -- epsilon for relative BM candidate threshold:
                   y in Y if D(x,y) <= (1+epsilon) * min d(x,y'),
                   default=10E-8 (for limited float precision).
        closest_outgroups -- use the closest outgroups corrected
                             for ancient duplications
        incongruence_threshold -- threshold for discarding
                                  outgroup species
        benchmark_file -- activate benchmarking in 'qinfer' and
                          specify the filename
        binary_path -- path to 'qinfer' binary (if not available
                       within path)
    """
    
    if not binary_path:
        qinfer_command = "qinfer"
    elif os.path.exists(binary_path):
        qinfer_command = binary_path
    else:
        raise FileNotFoundError("path to qinfer binary file '{}' does not exist".format(binary_path))
    
    output = -1
    
    command = [qinfer_command, matrix_filename, species_filename, tree_filename]
    
    if epsilon != -1:
        command.append( "--epsilon=" + str(epsilon) )
    if closest_outgroups:
        command.append( "--all-outgroups=" + str(incongruence_threshold) )
    if benchmark_file is not None:
        command.append( "--benchmark=" + benchmark_file )
    
    # call 'qinfer' and measure execution time
    start = time.time()
    
    try:
        output = subprocess.run(command, stdout=subprocess.PIPE)
    except:
        raise FileNotFoundError("calling qinfer failed")
        
    exec_time = time.time() - start
    
    if output == -1:
        raise Exception("no output from qinfer")
    
    bmg = parse_bmg_edges(output.stdout.decode(), scenario)
    
    return bmg, bmg.to_undirected(reciprocal=True), exec_time


def quartet_from_scenario(scenario, epsilon=-1,
                          closest_outgroups=False,
                          incongruence_threshold=0.2,
                          benchmark_file=None):
    """Compute BMG and RBMG from a distances matrix D using 'qinfer'.
    
    Keyword arguments:
        epsilon -- epsilon for relative BM candidate threshold:
                   y in Y if D(x,y) <= (1+epsilon) * min d(x,y'),
                   default=10E-8 (for limited float precision).
        closest_outgroups -- use the closest outgroups corrected
                             for ancient duplications
        incongruence_threshold -- threshold for discarding
                                  outgroup species
        benchmark_file -- activate benchmarking in 'qinfer' and
                          specify the filename
    """

    matrix_filename = "temp.phylip"
    species_filename = "temp_species.txt"
    tree_filename = "temp_tree.txt"
    
    matrix = scenario.get_distance_matrix()
    matrix_to_phylip(matrix_filename, scenario.genes, matrix)
    species_to_genes(species_filename, scenario)
    write_newick(tree_filename, scenario.S)
    
    bmg, rbmg, exec_time = quartet_qinfer(scenario,
                                          matrix_filename, species_filename, tree_filename,
                                          epsilon=epsilon,
                                          closest_outgroups=closest_outgroups,
                                          incongruence_threshold=incongruence_threshold,
                                          benchmark_file=benchmark_file)
    
    os.remove(matrix_filename)
    os.remove(species_filename)
    os.remove(tree_filename)
    
    return bmg, rbmg, exec_time
